@@ -1,4 +1,4 @@
-﻿"""
+"""
 Implements the "hardware" for tracing a function.
 
 Will perform the instructions and updates CPU registers, stack information, etc.
@@ -10,7 +10,11 @@ WARNING:
 from copy import deepcopy
 import collections
 import logging
+import warnings
+from typing import List, Tuple, Optional, Union
 
+import ida_frame
+import ida_funcs
 import idaapi
 import idc
 import ida_struct
@@ -21,8 +25,8 @@ from kordesii.utils.function_tracing.constants import *
 from kordesii.utils.function_tracing.memory import Memory
 from kordesii.utils.function_tracing.variables import VariableMap
 from kordesii.utils.function_tracing.operands import Operand, OperandLite
-from kordesii.utils.function_tracing.functions import FunctionSignature
-
+from kordesii.utils.function_tracing.functions import FunctionSignature, FunctionArg
+from kordesii.utils.function_tracing.objects import ObjectMap, Object, File, RegKey
 
 logger = logging.getLogger(__name__)
 
@@ -93,23 +97,27 @@ class JccContext(object):
         return self.condition_target_ea and self.condition_target_ea != ip
 
 
-# TODO: Create architecture specific Context types.
 class ProcessorContext(object):
     """
     Stores the context of the processor during execution.
 
+    :param emulator: Instance of Emulator to use during emulation.
     :param registers: Instance of an initialized RegisterMap object used to store register values
         for the given architecture.
     :param str instruction_pointer: Name of the register used to point to the current instruction
         being currently executed or to-be executed.
-    :param [str] stack_registers: List of register names used for handling the stack.
+    :param str stack_pointer: Name of the register used to hold the stack pointer.
     """
 
     # Must be set by inherited classes.
     ARCH_NAME = None  # Name of architecture as reported by disassembler.
     OPCODES = {}  # Map of opcode mnemonics to functions that emulate them.
 
-    def __init__(self, registers, instruction_pointer, stack_pointer, stack_registers=None):
+    # Cache for keeping track of instructions and their operand indexes.
+    _operand_indices = {}
+
+    def __init__(self, emulator, registers, instruction_pointer, stack_pointer):
+        self.emulator = emulator
         self.registers = registers
         self.jcccontext = JccContext()
         self.memory = Memory()
@@ -118,16 +126,23 @@ class ProcessorContext(object):
         self.memory_copies = collections.defaultdict(list)  # Keeps track of memory moves.
         self.bitness = utils.get_bits()
         self.byteness = self.bitness // 8
-        self.stack_registers = stack_registers or []
         self.variables = VariableMap(self)
+        self.objects = ObjectMap(self)
+        self.actions = []   # List of action objects (namedtuples)
+
+        # Function start address of a function we are currently hooking.
+        self.hooking_call = None
+
         self._sp = stack_pointer
         self._ip = instruction_pointer
+        self._sp_start = self.sp
 
     @classmethod
-    def from_arch(cls, arch_name=None):
+    def from_arch(cls, emulator, arch_name=None):
         """
         Factory method for initializing a ProcessorContext based on detected architecture.
 
+        :param emulator: Instance of Emulator to use during emulation.
         :param arch_name: Name of architecture to initializes (according to the disassembler)
                           Architecture is automatically detected if not provided.
 
@@ -140,7 +155,7 @@ class ProcessorContext(object):
 
         for subclass in cls.__subclasses__():
             if subclass.ARCH_NAME == arch_name:
-                return subclass()  # Subclasses shouldn't have any initialization parameters.
+                return subclass(emulator)
         raise NotImplementedError("Architecture not supported: {}".format(arch_name))
 
     def __deepcopy__(self, memo):
@@ -151,23 +166,27 @@ class ProcessorContext(object):
         copy = klass.__new__(klass)
         memo[id(self)] = copy
 
+        copy.emulator = self.emulator  # This is a reference, don't create a new instance.
+        copy.hooking_call = self.hooking_call
         copy.registers = deepcopy(self.registers, memo)
         copy.jcccontext = deepcopy(self.jcccontext, memo)
         copy.memory = deepcopy(self.memory, memo)
         copy.variables = deepcopy(self.variables, memo)
+        copy.objects = deepcopy(self.objects, memo)
+        copy.actions = list(self.actions)
         copy.func_calls = dict(self.func_calls)
         copy.executed_instructions = list(self.executed_instructions)
         copy.memory_copies = self.memory_copies.copy()
         copy.bitness = self.bitness
         copy.byteness = self.byteness
-        copy.stack_registers = self.stack_registers
         copy._sp = self._sp
         copy._ip = self._ip
+        copy._sp_start = self._sp_start
 
         return copy
 
     @property
-    def ip(self):
+    def ip(self) -> int:
         """Alias for retrieving instruction pointer."""
         return self.registers[self._ip]
 
@@ -177,7 +196,7 @@ class ProcessorContext(object):
         self.registers[self._ip] = value
 
     @property
-    def sp(self):
+    def sp(self) -> int:
         """Alias for retrieving stack pointer."""
         return self.registers[self._sp]
 
@@ -186,6 +205,17 @@ class ProcessorContext(object):
         """Alias for setting stack pointer."""
         self.registers[self._sp] = value
 
+    @property
+    def sp_diff(self) -> int:
+        """
+        The difference between the current stack pointer and the
+        stack pointer at the beginning of the function.
+
+        This helps with debugging since this number should match the number
+        shown in the IDA disassembly.
+        """
+        return self._sp_start - self.sp
+
     @property
     def prev_instruction(self):
         """That last instruction that was executed or None if no instructions have been executed."""
@@ -194,71 +224,184 @@ class ProcessorContext(object):
         else:
             return None
 
-    def execute(self, ip=None):
+    def execute_instruction_hooks(self, start, mnem, pre=True):
+        """
+        Executes instructions hooks for the given start
+        """
+        hooks = (
+            self.emulator.get_instruction_hooks(start, pre=pre)
+            + self.emulator.get_instruction_hooks(mnem, pre=pre)
+        )
+        if hooks:
+            operands = self.operands
+            for hook in hooks:
+                try:
+                    hook(self, start, mnem, operands)
+                except RuntimeError:
+                    raise  # Allow RuntimeError exceptions to be thrown.
+                except Exception as e:
+                    logger.debug("Failed to execute instruction hook with error: %s", e)
+
+    def execute(self, start=None, end=None, max_instructions=10000):
         """
         "Execute" the instruction at IP and store results in the context.
         The RIP/EIP register will be set to the value supplied in IP so that it is
         correct.
 
-        :param ip: instruction address to execute (defaults to currently set ip)
+        :param start: instruction address to start execution (defaults to currently set ip)
+        :param end: instruction to stop execution (not including)
+            (defaults to only run start)
+        :param max_instructions: Maximum number of instructions to execute before
+            raising an RuntimeError
+
+        :raises RuntimeError: If maximum number of instructions get hit.
         """
-        if not ip:
-            ip = self.ip
+        if not start:
+            start = self.ip
 
         # Set instruction pointer to where we are currently executing.
-        self.ip = ip
+        self.ip = start
+
+        # Extra processing if we are at the start of a function.
+        func_obj = ida_funcs.get_func(self.ip)
+        if func_obj.start_ea == self.ip:
+            # Reset the sp_start
+            self._sp_start = self.sp
+
+            # Add the passed in arguments to the variables map.
+            # (This also helps to standardize the argument names to "a*" instead of "arg_*")
+            for arg in self.passed_in_args:
+                addr = arg.addr
+                # TODO: Support variables from registers?
+                if addr is not None:
+                    if arg.is_stack:
+                        try:
+                            frame = ida_frame.get_frame(func_obj)
+                            if not frame:
+                                logger.warning(f"Failed to get frame for function argument: {repr(arg)}")
+                                continue
+
+                            # Getting member stack offset from name is more reliable then calculating
+                            # it from the address.
+                            member = ida_struct.get_member_by_name(frame, arg.name)
+                            if not member:
+                                logger.warning(f"Failed to get member for function argument: {repr(arg)}")
+                                continue
+
+                            self.variables.add(addr, frame_id=frame.id, stack_offset=member.soff)
+                        except ValueError:
+                            logger.warning(f"Failed to get stack information for function argument: {repr(arg)}")
+                    else:
+                        self.variables.add(addr)
+
+        # If end is provided, recursively run execute() until ip is end.
+        if end is not None:
+            count = max_instructions
+            prev_ecx = self.registers.ecx
+            prev_ecx_count = count
+            while self.ip != end:
+                if ida_ua.print_insn_mnem(self.ip) == 'retn':
+                    return
+                self.execute()
+                # TODO: Re-enable this feature after rigorous testing.
+                # # Dynamically allow more instructions to be executed if we detect we are in a loop
+                # # and it is making progress.
+                # # Ie. this will allow most while statements not to ding our max instruction quota.
+                # if self.registers.ecx and (self.registers.ecx < prev_ecx or prev_ecx == 0):
+                #     if prev_ecx:
+                #         count += min(prev_ecx_count - count, 1000)
+                #     prev_ecx = self.registers.ecx
+                #     prev_ecx_count = count
+                count -= 1
+                if not count:
+                    raise RuntimeError('Hit maximum number of instructions.')
+            return
 
         # Determine if a rep* instruction and add termination condition.
         term_condition = None
-        if idc.get_wide_byte(ip) in (0xF2, 0xF3):
-            insn = idc.GetDisasm(ip)  # IDA pro never has operands for rep opcodes.
+        if idc.get_wide_byte(start) in (0xF2, 0xF3):
+            insn = idc.GetDisasm(start)  # IDA pro never has operands for rep opcodes.
             if insn.startswith("rep "):
                 term_condition = lambda: self.registers.ecx == 0
+                term_condition.unconditional = True
             elif insn.startswith(("repe ", "repz ")):
                 term_condition = lambda: self.registers.ecx == 0 or self.registers.zf == 0
+                term_condition.unconditional = False
             elif insn.startswith(("repne ", "repnz ")):
                 term_condition = lambda: self.registers.ecx == 0 or self.registers.zf == 1
+                term_condition.unconditional = False
 
         # Emulate instruction.
-        mnem = idc.print_insn_mnem(ip)
-        operands = self.operands
+        mnem = ida_ua.print_insn_mnem(start)
+
+        # Log a header line for debug messages of this instruction.
+        # This is simpler and faster then trying to include the information at each log line
+        logger.debug("[0x%X %03X] :: %s", start, self.sp_diff, mnem)
+
+        # Run any pre-hooks first.
+        self.execute_instruction_hooks(start, mnem, pre=True)
+
         instruction = self.OPCODES.get(mnem)
         if instruction:
+            operands = self.operands
+
             try:
                 if term_condition:
+                    if self.emulator.disabled_rep:
+                        logger.debug("Ignoring rep instructions: DISABLED.")
+
                     # As a safety measure, don't allow rep instructions to surpass
                     # our max memory read limit.
-                    if self.registers.ecx > self.memory.MAX_MEM_READ:
+                    # Only do this check if the terminating condition is unconditional, otherwise
+                    # this number usually big because it expects zf to be toggled.
+                    elif term_condition.unconditional and self.registers.ecx > self.memory.MAX_MEM_READ:
                         logger.warning(
-                            "0x{:08X} :: Emulation attempted to read {} instruction {} times. "
-                            "Ignoring instruction.".format(ip, mnem, self.registers.ecx)
+                            "Emulation attempted to read %s instruction %d times. "
+                            "Ignoring instruction.", mnem, self.registers.ecx
                         )
                     else:
-                        logger.debug("Emulating {} instruction {} times.".format(mnem, self.registers.ecx))
+                        logger.debug("Emulating %s instruction %d times.", mnem, self.registers.ecx)
+                        count = 0
                         while not term_condition():
-                            instruction(self, ip, mnem, operands)
+                            instruction(self, start, mnem, operands)
                             self.registers.ecx -= 1
+                            # Stop if we are iterating too much.
+                            count += 1
+                            if count > self.memory.MAX_MEM_READ:
+                                logger.warning("Looped too many times, exiting prematurely.")
+                                break
                 else:
-                    instruction(self, ip, mnem, operands)
+                    instruction(self, start, mnem, operands)
             except Exception:
-                logger.exception("Failed to execute address 0x{:X}: {}".format(ip, idc.GetDisasm(ip)))
+                logger.exception("Failed to execute address 0x%X: %s", start, idc.GetDisasm(start))
         else:
-            logger.debug("{} instruction not implemented.".format(mnem))
+            logger.debug("%s instruction not implemented.", mnem)
 
         # Record executed instruction.
-        self.executed_instructions.append(ip)
+        self.executed_instructions.append(start)
+
+        # Run any post-hooks
+        self.execute_instruction_hooks(start, mnem, pre=False)
+
+        # Add a blank space to help visually separate logs for each instruction.
+        logger.debug(" ")
 
         # After execution, set instruction pointer to next instruction assuming
         # standard code flow and if no jump was made.
-        if self.ip == ip:
-            self.ip = idc.next_head(ip)
+        if self.ip == start:
+            self.ip = idc.next_head(start)
 
-    def get_call_history(self, func_name):
+    def get_call_history(self, func_name_or_ea) -> List[Tuple[int, List]]:
         """
         Returns the call history for a specific function name.
 
-        :returns: List of tulples containing: (ea of call, list of function arguments)
+        :returns: List of tuples containing: (ea of call, list of function arguments)
         """
+        if isinstance(func_name_or_ea, str):
+            func_name = func_name_or_ea
+        else:
+            ea = func_name_or_ea
+            func_name = utils.get_function_name(ea)
         return [(ea, args) for ea, (_func_name, args) in list(self.func_calls.items()) if _func_name == func_name]
 
     def prep_for_branch(self, bb_start_ea):
@@ -266,7 +409,7 @@ class ProcessorContext(object):
         Modify this current context in preparation for a specific path.
         """
         if self.jcccontext.is_alt_branch(bb_start_ea):
-            logger.debug("Modifying context for branch at 0x{:X}".format(bb_start_ea))
+            logger.debug("Modifying context for branch at 0x%08X", bb_start_ea)
             # Set the destination operand relative to the current context
             # to a valid value that makes this branch true.
             dst_opnd = self.jcccontext.alt_branch_data_dst
@@ -275,7 +418,7 @@ class ProcessorContext(object):
 
         self.jcccontext = JccContext()
 
-    def get_operands(self, ip=None):
+    def get_operands(self, ip=None) -> List[Operand]:
         """
         Gets the Operand objects of all operands in the current instruction and returns them in a list.
 
@@ -286,22 +429,30 @@ class ProcessorContext(object):
         if ip is None:
             ip = self.ip
 
-        operands = []
-        cmd = ida_ua.insn_t()
-        # NOTE: We can't trust the instruction length returned by decode_ins.
-        ida_ua.decode_insn(cmd, ip)
-        for idx, op in enumerate(cmd.ops):
-            operand = Operand(self, ip, idx)
-            # IDA will sometimes create hidden or "fake" operands.
-            # These are there to represent things like an implicit EAX register.
-            # To help avoid confusion to the opcode developer, these fake operands will not be included.
-            if not operand.is_hidden:
-                operands.append(operand)
+        # Calling insn_t() and decode_insn() is somewhat expensive and this function gets called a LOT,
+        # so we are going to cache the operand indices.
+        try:
+            indices = self._operand_indices[ip]
+        except KeyError:
+            indices = []
+            insn = ida_ua.insn_t()
+            # NOTE: We can't trust the instruction length returned by decode_ins.
+            ida_ua.decode_insn(insn, ip)
+            for idx, op in enumerate(insn.ops):
+                if op.type == ida_ua.o_void:
+                    break  # no more operands
 
-            if operand.is_void:
-                break  # no more operands
+                # IDA will sometimes create hidden or "fake" operands.
+                # These are there to represent things like an implicit EAX register.
+                # To help avoid confusion to the opcode developer, these fake operands will not be included.
+                # TODO: Checking shown() may not work as expected.
+                #   If things explode, go back to checking operand.is_hidden
+                if op.shown():
+                    indices.append((idx, op.type))
 
-        return operands
+            self._operand_indices[ip] = indices
+
+        return [Operand(self, ip, idx, _type=type) for idx, type in indices]
 
     @property
     def operands(self):
@@ -318,7 +469,7 @@ class ProcessorContext(object):
 
         :return int: value contained in specified register as int
         """
-        return self.registers[reg]
+        return self.registers[reg.lower()]
 
     def reg_write(self, reg, val):
         """
@@ -328,7 +479,7 @@ class ProcessorContext(object):
 
         :param int val: value to be written to register as an int of width of the register (will be truncated as necessary)
         """
-        self.registers[reg] = val
+        self.registers[reg.lower()] = val
 
     def mem_alloc(self, size):
         """
@@ -440,6 +591,7 @@ class ProcessorContext(object):
     def mem_find_in_heap(self, value):
         return self.memory.find_in_heap(value)
 
+    # TODO: Move this into Memory class and automatically decode values.
     def read_data(self, addr, size=None, data_type=None):
         """
         Reads memory at the specified address, of the specified size and convert
@@ -448,7 +600,7 @@ class ProcessorContext(object):
         :param int addr: address to read data from
         :param int size: size of data to read
         :param data_type: type of data to be extracted
-            (default to byte string is size provided or C string if not)
+            (default to BYTE_STRING is size provided or STRING if not)
         """
         if not data_type:
             data_type = STRING if size is None else BYTE_STRING
@@ -488,7 +640,59 @@ class ProcessorContext(object):
 
         raise ValueError("Invalid data_type: {!r}".format(data_type))
 
-    def get_function_signature(self, func_ea=None, force=False) -> FunctionSignature:
+    def write_data(self, addr, value, data_type=None):
+        """
+        Writes memory at the specified address after converting the value
+        into data based on the specified data type.
+
+        :param int addr: address to write data to
+        :param value: integer or byte string to write
+        :param data_type: type of data to convert value from.
+            (defaults to BYTE_STRING, STRING, or DWORD based on input data)
+        """
+        if not data_type:
+            if isinstance(value, str):
+                data_type = STRING
+            elif isinstance(value, bytes):
+                data_type = BYTE_STRING
+            elif isinstance(value, int):
+                data_type = DWORD
+            else:
+                raise ValueError(f"Invalid data type: {type(value)}")
+
+        if data_type == BYTE_STRING:
+            data = value
+
+        elif data_type == STRING:
+            data = value
+            if isinstance(data, str):
+                data = data.encode("utf8")
+            data += b"\0"
+
+        elif data_type == WIDE_STRING:
+            data = value
+            if isinstance(data, str):
+                data = data.encode("utf-16-le")
+            data += b"\0\0"
+
+        elif data_type == BYTE:
+            data = bytes([value])
+
+        elif data_type == WORD:
+            data = utils.struct_pack(value, width=2)
+
+        elif data_type == DWORD:
+            data = utils.struct_pack(value, width=4)
+
+        elif data_type == QWORD:
+            data = utils.struct_pack(value, width=8)
+
+        else:
+            raise ValueError(f"Invalid data_type: {repr(data_type)}")
+
+        self.mem_write(addr, data)
+
+    def get_function_signature(self, func_ea=None, force=False, num_args=None) -> Optional[FunctionSignature]:
         """
         Returns the function signature of the given func_ea with argument values pulled
         from this context.
@@ -498,18 +702,32 @@ class ProcessorContext(object):
         :param bool force: Whether to force a function signature using cdecl calling
             convention and no arguments if we fail to generate the signature.
             (Useful when trying to declare a function that was dynamically created in a register)
-        :return: FunctionSignature object
+        :param int num_args: Force a specific number of arguments in the signature.
+            If not provided, number of arguments is determined by the disassembler.
+            Extra arguments not defined by the disassembler are assumed to be 'int' type.
+            Avoid using num_args and adjust the returned FunctionSignature manually
+            if more customization is needed.
+            (NOTE: The function signature will be forced on failure if this is set.)
+
+        :return: FunctionSignature object or None if not applicable
 
         :raises RuntimeError: If a function signature could not be created from given ea.
+        :raises ValueError: If num_args is negative
         """
         # If func_ea is not given, assume we are using the first operand from a call instruction.
         if not func_ea:
+            if not self.operands:
+                return None
             operand = self.operands[0]
             # function pointer can be a memory reference or immediate.
             func_ea = operand.addr or operand.value
+        else:
+            operand = None
+
+        force = force or num_args is not None
 
         try:
-            return FunctionSignature(self, func_ea)
+            func_sig = FunctionSignature(self, func_ea, operand=operand)
         except RuntimeError as e:
             # If we fail to get a function signature but force is set, set the type to
             # cdecl with no arguments.
@@ -519,30 +737,10 @@ class ProcessorContext(object):
                     "Forcing signature with assumed cdecl calling convention.".format(func_ea, e)
                 )
                 idc.SetType(func_ea, "int __cdecl no_name();")
-                return FunctionSignature(self, func_ea)
+                func_sig = FunctionSignature(self, func_ea)
             else:
                 raise
 
-    def get_function_args(self, func_ea=None, num_args=None):
-        """
-        Returns the function argument values for this context based on the
-        given function.
-
-        >>> cpu_context = ProcessorContext()
-        >>> args = cpu_context.get_function_args(0x180011772)
-
-        :param int func_ea: Ea of the function to pull a signature from.
-        :param int num_args: Force a specific number of arguments.
-            If not provided, number of arguments is determined by the disassembler.
-            Extra arguments not defined by the disassembler are assumed to be 'int' type.
-            Use get_function_signature() and adjust the FunctionSignature manually
-            if more customization is needed.
-            (NOTE: The function signature will be forced on failure if this is set.)
-
-        :returns: list of function arguments
-        """
-        func_sig = self.get_function_signature(func_ea, force=num_args is not None)
-
         if num_args is not None:
             if num_args < 0:
                 raise ValueError("num_args is negative")
@@ -552,4 +750,111 @@ class ProcessorContext(object):
             elif len(arg_types) < num_args:
                 func_sig.arg_types = arg_types + ("int",) * (num_args - len(arg_types))
 
-        return [arg.value for arg in func_sig.args]
+        return func_sig
+
+    def get_function_arg_objects(self, func_ea=None, num_args=None) -> List[FunctionArg]:
+        """
+        Returns the FunctionArg objects for this context based on the
+        given function.
+
+        >>> cpu_context = ProcessorContext()
+        >>> args = cpu_context.get_function_arg_objects(0x180011772)
+
+        :param int func_ea: Ea of the function to pull a signature from.
+        :param int num_args: Force a specific number of arguments.
+            If not provided, number of arguments is determined by the disassembler.
+            Extra arguments not defined by the disassembler are assumed to be 'int' type.
+            Use get_function_signature() and adjust the FunctionSignature manually
+            if more customization is needed.
+            (NOTE: The function signature will be forced on failure if this is set.)
+
+        :returns: list of FunctionArg objects
+        """
+        func_sig = self.get_function_signature(func_ea, num_args=num_args)
+        if not func_sig:
+            return []
+
+        return func_sig.args
+
+    # TODO: Replace this function with get_function_arg_objects()
+    def get_function_args(self, func_ea=None, num_args=None) -> List[int]:
+        """
+        Returns the FunctionArg values for this context based on the given function.
+        """
+        args = self.get_function_arg_objects(func_ea=func_ea, num_args=num_args)
+        return [arg.value for arg in args]
+
+    @property
+    def function_args(self) -> List[FunctionArg]:
+        """
+        The function arguments currently set based on the function in the first operand.
+        """
+        return self.get_function_arg_objects()
+
+    @property
+    def passed_in_args(self) -> List[FunctionArg]:
+        """
+        The function arguments for the current function.
+        """
+        func = ida_funcs.get_func(self.ip)
+        return self.get_function_arg_objects(func.start_ea)
+
+    @property
+    def files(self) -> List[File]:
+        """
+        The opened files for this context.
+        """
+        return [obj for obj in self.objects if isinstance(obj, File)]
+
+    def open_file(self, path: str, mode: str = None) -> File:
+        """
+        Adds opened file for tracking.
+        """
+        if not path:
+            path = f"0x{self.ip:08x}.bin"
+
+        # First see if the file already exists.
+        for file in self.files:
+            if file.path == path:
+                if mode:
+                    file.mode = mode  # update mode
+                # Undo closed and delete indicators.
+                file.closed = False
+                file.deleted = False
+                return file
+
+        # Create a new file and store in object map.
+        file = File(path=path, mode=mode)
+        self.objects.add(file)
+        logger.debug("Opened file: %s", file.path)
+        return file
+
+    def get_file(self, handle_or_path: Union[int, str], default=None) -> File:
+        """
+        Gets a file by handle or path. Returns default if not existent.
+
+        :raises TypeError: if handle points and object that is not a File.
+        """
+        if isinstance(handle_or_path, int):
+            try:
+                handle = handle_or_path
+                obj = self.objects[handle]
+                if not isinstance(obj, File):
+                    raise TypeError(f"Expected handle {hex(handle)} to point to a File, got {type(obj)}")
+                return obj
+            except KeyError:
+                return default
+
+        else:
+            path = handle_or_path
+            for file in self.files:
+                if file.path == path:
+                    return file
+            return default
+
+    @property
+    def regkeys(self) -> List[RegKey]:
+        """
+        The opened registry keys for this context.
+        """
+        return [obj for obj in self.objects if isinstance(obj, RegKey)]
