@@ -8,6 +8,7 @@ import string
 import struct
 
 import ida_allins
+import ida_ida
 import ida_idp
 import ida_nalt
 import ida_typeinf
@@ -16,6 +17,7 @@ import idautils
 import idc
 
 from .exceptions import FunctionTracingError
+from .operands import Operand
 
 logger = logging.getLogger(__name__)
 
@@ -99,7 +101,7 @@ def get_byte_width(value):
         return 8
 
 
-BIG_ENDIAN = idaapi.cvar.inf.is_be()
+BIG_ENDIAN = ida_ida.inf_is_be()
 
 # Table used by struct_unpack and struct_pack functions
 struct_unpack_table = {
@@ -250,13 +252,15 @@ def get_bits():
 _func_types = set()
 
 
-def get_function_data(offset):
+def get_function_data(offset, operand: Operand = None):
     """
     Obtain a idaapi.func_type_data_t object for the function with the provided start EA.
 
     :param int offset: start EA of function
+    :param operand: operand containing function address in it's value.
+        This can be provided when function is dynamically generated at runtime. (e.g. call eax)
 
-    :return: idaapi.func_type_data_t object
+    :return: ida_typeinf.func_type_data_t object, ida_typeinf.tinfo_t object
 
     :raise RuntimeError: if func_type_data_t object cannot be obtained
     """
@@ -328,13 +332,22 @@ def get_function_data(offset):
                     raise RuntimeError("failed to obtain tinfo_t object for offset 0x{:X}".format(offset))
 
     funcdata = ida_typeinf.func_type_data_t()
-    if not tif.get_func_details(funcdata):
-        raise RuntimeError("failed to obtain func_type_data_t object for offset 0x{:X}".format(offset))
+    success = tif.get_func_details(funcdata)
+    if success:
+        # record that we have processed this function before. (and that we can grab it from the offset)
+        _func_types.add(offset)
+        return funcdata, tif
 
-    # record that we have processed this function before.
-    _func_types.add(offset)
+    # If we have still failed, we have one more trick under our sleeve.
+    # Try to pull the type information from the operand of the call instruction.
+    # This could be set if the function has been dynamically created.
+    if operand:
+        tif = operand._tif
+        success = tif.get_func_details(funcdata)
+        if success:
+            return funcdata, tif
 
-    return funcdata
+    raise RuntimeError("failed to obtain func_type_data_t object for offset 0x{:X}".format(offset))
 
 
 def get_function_name(func_ea):
